/*
CubeStat : Estimates the size of the data along each level in the dimension hierarchy
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package cube;

import java.util.*;
import java.io.*;
import java.sql.*;

import util.*;

public class CubeStat {
    Vector tempVec = new Vector();
    RelCubeC cube;
    int[] estSizes;
    
    public CubeStat(RelCubeC c) {
	cube = c;
    }
    public Object randomDescendant(AttrDesc parDesc, Object parVal, AttrDesc childDesc) {
	try {
	    ResultSet rs = cube.db().executeQuery("select distinct " + cube.oidFld(childDesc.dim(), childDesc.level()) + " from " + cube.dimTable(childDesc.dim()) + " where " + cube.oidFld(parDesc.dim(), parDesc.level()) + " = " + parVal);
	    rs.next();
	    Object selchild = rs.getObject(1);
	    int pos = 1;
	    while (rs.next()) {
		Object tmp = rs.getObject(1);
		pos++;
		if (Math.random()*pos < 1)
		    selchild = tmp;
	    }
	    DB.cursorClose(rs);
	    return selchild;
	} catch (Exception e) {
	    e.printStackTrace();
	}
	return null;
    }

    int getPos(Term t)
    {
	int pos = 0;
	int offset = 1;
	for (int i = 0; i < t.numAttrs(); i++) {
	    int lvl = cube.numLevels(t.desc(i).dim());
	    for (int j = i; (j < t.numAttrs()) && (t.desc(j).dim() == t.desc(i).dim()); i=j, j++) {
		if (t.constant(j) || t.range(j))
		    lvl = t.desc(j).level();
	    }
	    
	    pos += offset*lvl;
	    offset *= (cube.numLevels(t.desc(i).dim())+1);
	}
	return pos;
    }

    void setTermFromPos(Term t, int pos)
    {
	int offset = 1;
	for (int i = 0; i < t.numAttrs(); i++) {
	    offset *= (cube.numLevels(t.desc(i).dim())+1);
	    int lvl = (pos % offset);
	    pos = pos/offset;
	    for (int j = i; (j < t.numAttrs()) && (t.desc(j).dim() == t.desc(i).dim()); i=j, j++) {
		if (lvl > t.desc(j).level())
		    t.setVal(Term.AGGREGATED, j);
		else
		    t.setVal(Term.RANGE, j);
	    }
	}
    }

    public int sizeEstimate(Term t) {
	int cvr = 1;  
	float selectivity = 1;
	for (int i = 0; i < t.numAttrs(); i++) {
	    int lvl = -1;
	    for (int j = i; (j < t.numAttrs()) && (t.desc(j).dim() == t.desc(i).dim()); i=j, j++) {
		if (t.constant(j))
		    lvl = t.desc(j).level();
	    }
	    if (lvl >= 0) 
		selectivity *= cube.numMembers(t.desc(i).dim(), lvl);
	}
	return (int)(getSize(getPos(t))/selectivity);
    }
    public int getSize(int pos) {
	return estSizes[pos];
    }
  public int numDescendants(AttrDesc parDesc, Object parVal, AttrDesc childDesc) {
    if (parDesc == null)
      return cube.numMembers(childDesc.dim(), childDesc.level());
    return cube.numMembers(childDesc.dim(), childDesc.level())/cube.numMembers(parDesc.dim(), parDesc.level());
  }


  int getMaxSize(Term t) {
    float sz = 1;
    for (int i = 0; i < t.numAttrs(); i++) {
      if (t.aggregated(i))
	continue;
      int lvl = cube.numLevels(t.desc(i).dim());
      for (int j = i; (j < t.numAttrs()) && (t.desc(j).dim() == t.desc(i).dim()); i=j, j++) {
	if (t.range(j))
	  lvl = t.desc(j).level();
      }
      sz *=  cube.numMembers(t.desc(i).dim(),lvl); 
    }
    return (int)Math.min(sz, (float)(Integer.MAX_VALUE));
  }


  class Estimate {
      int numNodes;
      HashIndices[] estArray;
      int NUM_BITMAPS= 64;
    
      int totalstr = 0;
      Term[] mask;
      int[] str;
      
      Term newMask(Term t, int k) {
	  Term tt = new Term(t);
	  setTermFromPos(tt, k);
	  return tt;
      }
      private int[] convert(Term t, Term mask) {
	  for (int i = 0; i < mask.numAttrs(); i++) {
	      str[t.desc(i).dim()] = Term.AGGREGATED.hashCode();
	      for (int j = i; (j < t.numAttrs()) && (t.desc(j).dim() == t.desc(i).dim()); i=j, j++) {
		  if (mask.range(j))
		      str[t.desc(i).dim()] = t.val(j).hashCode();
	      }
	  }
	  return str;
      }
      public Estimate(Cube cube) {
	  // initialize the hash-structures for each node
	  numNodes = 1;
	  for (int i = 0; i < cube.numDims(); i++)
	      numNodes *= (cube.numLevels(i)+1);
 
	  str = new int[cube.numDims()];
	  estArray = new HashIndices[numNodes];
	  mask[0] = new Term(MdUtils.getColDesc(cube));
	  for (int k = 1; k < numNodes; k++) {
	      mask[k] = newMask(mask[0], k);
	      estArray[k].init(getMaxSize(mask[k]), NUM_BITMAPS);
	  }
      }
      void add(Term t) {
	  totalstr++;
	  // for each node, generate the mask and add to hash structure
	  for (int nodeIndex = numNodes-1; nodeIndex > 0; nodeIndex--) {
	      estArray[nodeIndex].add(convert(t, mask[nodeIndex]), cube.numDims());
	  }
      }
      
      int[] setEstimate()
      {
	  int[] sizes = new int[numNodes];
	  for (int nodeIndex = numNodes-1; nodeIndex > 0; nodeIndex--) {
	      int nstr = Math.min(estArray[nodeIndex].estimate(),totalstr);
	      sizes[nodeIndex] = nstr;
	  }
	  sizes[0] = 1;
	  return sizes;
      }
  };
  
  void EstimateSizes(Cube cube, Iter iter) 
  {
    Estimate estimate = new Estimate(cube);
    while (iter.moreTuples()) {
      estimate.add(iter.getTuple());
    }
    estSizes = estimate.setEstimate();
  }
}

