/*
Diff : Outputs the differences between two Terms in a compact form
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// change the tuple description from static to pointer..
//
package diff;


import cube.*;
import util.*;
import java.util.*;

public class Diff {
    Cube _cube;
    DiffParams _pars;
    TermDesc coldesc;

    public Diff(Cube cube, DiffParams pars) {
	_cube = cube;
	_pars = pars;
	coldesc = new TermDesc(_cube.numAttrs());
	for (int i = 0, r = 0; i < _cube.numDims(); i++) {
	    for (int j = _cube.numLevels(i)-1; j >= 0; j--) {
		coldesc.desc(r++).init(i,j,_cube.attrName(i,j));
	    }
	}
    }
    public void setParams(String pars) {
	_pars = new DiffParams(pars);
    }
    public Result getDiff(String attrNames, String attrVals1, String attrVals2, String aggAttrs, String aggVals) throws Exception
    {
	String[] names = MdUtils.getArray(attrNames);
	Term t1 = MdUtils.getTermMapping(_cube,coldesc,names
				       ,MdUtils.getArray(attrVals1));
	Term t2 = MdUtils.getTermMapping(_cube,coldesc,names
					 ,MdUtils.getArray(attrVals2));
	Term aggLevel = null;
	if (aggAttrs != null) 
	    aggLevel = MdUtils.getTermMapping(_cube, coldesc,
					      MdUtils.getArray(aggAttrs),
					      MdUtils.getArray(aggVals));
	DiffResult res = getDiffInt(t1, t2,aggLevel);
	return new FullResult(res, t1, t2, _cube);
	// new ResultStore(res, _cube);
    }

    public FullResult getDiff(Vector attrNames1, Vector attrVals1, Vector attrNames2, Vector attrVals2) throws Exception
    {
	Term t1 = MdUtils.getMTermMapping(_cube,coldesc,attrNames1, attrVals1);
	Term t2 = MdUtils.getMTermMapping(_cube,coldesc,attrNames2,attrVals2);
	
	DiffResult res = getDiffInt(t1, t2, null);
	return new FullResult(res, t1, t2, _cube);
    }
    public Result getDiff(Result r) throws Exception {
	if ((r == null) || (r.numTups() == 0))
	    return null;
	FullResult res = (FullResult)r;
	Term t1 = res.firstTerm();
	Term t2 = res.secondTerm();

	Iter iter = getCubeIter(t1, t2, res.tup(0).termDescr());
	FindDiff fd =  new FindDiff(iter, _pars, res, null);
	DiffResult dres = fd.finalResult();
	dres.output();
	res.addTuples(dres,_cube);
	return res;
    }
    public Result getDiff(Term t1, Term t2, Term aggLevel) throws Exception {
	System.out.println(t1.toString());
	System.out.println(t2.toString());
	t1.validate(_cube);
	t2.validate(_cube);
	return new FullResult(getDiffInt(t1,t2,aggLevel),t1,t2,_cube);
    }

    Iter getCubeIter(Term t1, Term t2, TermDesc tdesc) throws Exception {
	QueryHandle qh = _cube.concatCubes(t1, t2,tdesc, 
			  new MeasureExpr("sum("+_cube.measure(0)+")","v0"),
			  new MeasureExpr("sum("+_cube.measure(0)+")","v1"),
					   MdUtils.defaultArray(t1));
	return new RelDataIter(_cube.executeQuery(qh), tdesc);
			  
    }
    // aggLevel refers to the depth to which you want to drill down.
    public DiffResult getDiffInt(Term t1, Term t2, Term aggLevel) throws Exception {
	
	System.out.println(t1.toString());
	System.out.println(t2.toString());
	if (t1.isEqual(t2, null)) {
	    System.out.println("Terms are the same");
	    return null;
	}
	for (int i = 0; i < t1.numAttrs(); i++) {
	    if (t1.constant(i) && t2.constant(i)) {
		if (!t1.attrVal(i).same(t2.attrVal(i)) && (t1.desc(i).level() > 0)) {
		    // we only want to expand common dimensions
		    for (int j = i+1; (j < t1.numAttrs()) && 
			     (t1.desc(j).dim() == t1.desc(i).dim());j++,i++);
		}
		continue;
	    }
	    if ((aggLevel != null) && !aggLevel.range(i))
		continue;
	    if (t1.aggregated(i) && t2.aggregated(i)) {
		t1.setVal(Term.RANGE, i);
		t2.setVal(Term.RANGE, i);
		continue;
	    }
	    throw new Exception("DIFF: data format error: both terms should be either constant or aggregated");
	}
	TermDesc tdesc = MdUtils.getTermDescForRange(t1);
	if ((tdesc == null) || (tdesc.size() == 0))
	    return null;
	FindDiff fd;
	if (! _pars.exhaustive()) {
	    Iter iter = getCubeIter(t1,t2,tdesc);
	    fd =  new FindDiff(iter, _pars, null, null);
	} else {
	    IterGenC itergen = new IterGenC(_cube, t1, t2, tdesc);
	    fd =  new FindDiff(itergen, _pars, null);
	}
	DiffResult res = fd.finalResult();
	//	res.output();
	return res;
    }
    // 0 -- provider name: DB2/ORACLE/MSOLOP etc
    // 1 -- cube name
    // 2,3 -- usr,passwd
    // 4 -- parameters
    // 5,6,7 -- attributes, the two values to compare --- comma separated.
    // Example:
    // java diff.Diff DB2 data/stud dbuser dbpasswd "exhaustive;N=20" 
    // "Gender" "F" "M" 
    public static void main(String argv[]) {
	try {
	    DiffParams pars = new DiffParams(argv[4]);
	    Cube cube = MdUtils.parseInputPars(argv);
	    Diff diff = new Diff(cube, pars);
	    if (argv.length > 8) {
		diff.getDiff(argv[5], argv[6], argv[7],argv[8],argv[9]).output();
	    } else if (argv.length > 5) {
		diff.getDiff(argv[5], argv[6], argv[7],null,null).output();
	    }
	} catch(Exception e) {
	    e.printStackTrace();
	}
    }
}
