/*
DiffParams : Stores parameters specific to the DIFF operator
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package diff;
import java.util.*;
import util.*;
public class DiffParams extends Params
{
    boolean _fixedN = true; // solution is for fixed N or fixed error?
    public int _N = 10;
    double _maxError = 1.0;
    public int _maxR = 4;
    public float _ratios[] = null;
    public float _ratio = 1;
    boolean _debug = false;
    public boolean _exhaustive = false;    
    public Alg _alg = new KLDist();
    
    public float[] ratios() {return _ratios;}
    public int sign() {return 1;}
    public float ratio() {return _ratio;}
    public int N()  {return _N;}
    public float diff() {return 0;}
    public int maxR() {return _maxR;}
    public void setDebugOn() {_debug = true;}
    public boolean debug() {return _debug;}
    public boolean exhaustive() {return _exhaustive;}
    public Alg errorFunction() {return _alg;}
    public boolean fixedN() {return _fixedN;}
    public double maxError() {return _maxError;}
 
    public DiffParams() {super();}
    public DiffParams(String args) {
	super(args);
	if (args == null) 
	    return;
	StringTokenizer tok = new StringTokenizer(args, ";");
	while (tok.hasMoreTokens()) {
	    String anArg = tok.nextToken();
	    if (anArg.startsWith("N=")) 
		_N = Integer.valueOf(anArg.substring(2)).intValue();
	    if (anArg.startsWith("NR=")) 
		_maxR = Integer.valueOf(anArg.substring(3)).intValue();
	    if (anArg.equalsIgnoreCase("debugDiff"))
		setDebugOn();
	    if (anArg.startsWith("exhaust"))
		_exhaustive = Boolean.valueOf(anArg.substring(8)).booleanValue();
	    if (anArg.startsWith("R="))
		_ratio = Float.valueOf(anArg.substring(2)).floatValue();
	    if (anArg.startsWith("maxError=")) {
		_fixedN = false;
		_maxError = Double.valueOf(anArg.substring(9)).doubleValue();
	    }
	    if (anArg.startsWith("ErrFunc=")) {
		String errFunc = anArg.substring(8);
		if (errFunc.equalsIgnoreCase("Boolean")) {
		    System.out.println("Using boolean error function..");
		    _alg = new BooleanError();
//		} else if (errFunc.equalsIgnoreCase("Multiple")) {
//		    System.out.println("Using multiple error function..");
//		    _alg = new MultError();
		} else 	if (errFunc.equalsIgnoreCase("Weighted log-ratio")) {
		    System.out.println("Using weighted log-ratio function..");
		    _alg = new KLDist();
		} else if (errFunc.equalsIgnoreCase("SquareError")) {
		    System.out.println("Using square error function..");
		    _alg = new Alg();
		} else 	if (errFunc.equalsIgnoreCase("Unweighted log-ratio")) {
		    System.out.println("Using unweighted log-ratio function..");
		    _alg = new LogRatio();
		} 
	    }
	}
    }
};


