/*
KLDist : Calculates the error as the KL distance between two tuples
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


package diff;

public class KLDist extends Alg {
    //    public AlgType type()  {return POISSON;}
    public int numDerived()  {return 5;}
    public String derivedCols()  {return ", f1, f2";}
    public String derivedCalcs()  {
	return  ", sum(case when v2 = 0 then 0 else v2*log(v2/" + Alg.msgValSub() + ") end) as f1 ";
    }
    public String filterClause (double ratio)  {
	double descLen = avgDataDescLen();
	double lratio = Math.log(ratio);
	String sratio = String.valueOf(ratio);
	return "(f1 + c1*" + sratio + " - c2*" + String.valueOf(lratio) + ")> 2*numNulls*" + String.valueOf(descLen);
    }
    public void calcDerivedCols(double _v1, double _v2, double _dvals[])
    {
	if ((_v1 > Double.MIN_VALUE) && (_v2 > Double.MIN_VALUE)) {
	    double logr = Math.log(_v2/_v1);
	    _dvals[0] = _v2*logr;
	    _dvals[1] = _v1*logr;
	    _dvals[2] = _dvals[3] = 0;
	    _dvals[4] = _v2*Math.log(_v2);
	    return;
	} 
	_dvals[0] = _dvals[1] = 0;
	if (_v1 <= Double.MIN_VALUE) {
	    _dvals[2] = _v2; 
	    _dvals[3] = 0;
	    _dvals[4] = _v2*Math.log(_v2);
	    return;
	} 
	_dvals[0] = _dvals[1] = 0;
	_dvals[3] = _v1;
	_dvals[2] = 0;
	_dvals[4] = 0;
	return;
    }

    public double errorRatio(double _v1, double _v2, int cnt, double _dvals[], double r) {
	if ((_v2 <= Double.MIN_VALUE) || (_v1 <= Double.MIN_VALUE)) 
	    return 0;
	return error(_v1,_v2,cnt,_dvals,r)/(_dvals[4]); 
    }

    double smallRatio(double r) {
	double rr = Math.min(r-SMALL_RATIO, SMALL_RATIO); 
	if (rr > Double.MIN_VALUE)
	    return rr;
	return rr/2;
    }
    public double error(double _v1, double _v2, int cnt, double _dvals[], double r) {
	// find symmetric KL distance, blog(b/ar) + ar*log(ar/b) ==
	// b(log(b/ar) + ar/blog(ar/b)) = b*log(ar/b)*(ar/b-1) = b*log(rr)*(rr-1)
	double b =  _v2;
	double a =  _v1;

	if ((a <= Double.MIN_VALUE) && (b <= Double.MIN_VALUE)) 
	    // empty tuple..
	    return 0;
	if ((a <= Double.MIN_VALUE) && (r == LARGE_RATIO))
	    // aggregation of missing values.
	    return 0;
	if ((b <= Double.MIN_VALUE) && (r == 0))
	    // aggregation of missing values.
	    return 0;
	double logr = Math.log(r);
	if (r <= Double.MIN_VALUE)
	    logr = Math.log(Math.min(SMALL_RATIO, b/a));	
	double p = _dvals[0] 
	    + _dvals[2]*Math.log(Math.max(LARGE_RATIO, r+SMALL_RATIO))   
	    - b*logr 
	    - r*_dvals[1] 
	    - r*_dvals[3]*Math.log(smallRatio(r))
	    + a*r*logr;
	return p;
    }
};




