/*
LogRatio : Calculating error based on an unweighted log-ratio function
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package diff;

public class LogRatio extends Alg {
    public int numDerived()  {return 4;}
    public void calcDerivedCols(double _v1, double _v2, double _dvals[])
    {
	if ((_v1 > Double.MIN_VALUE) && (_v2 > Double.MIN_VALUE)) {
	    double logr = Math.log(_v2/_v1);
	    _dvals[0] = logr*logr;
	    _dvals[1] = logr;
	    return;
	} 
	if (_v1 <= Double.MIN_VALUE) {
	    _dvals[2] = 1; 
	    return;
	} 
	_dvals[3] = 1;
    }

    public double errorRatio(double _v1, double _v2, int cnt, double _dvals[], double r) {
	if ((_v2 <= Double.MIN_VALUE) || (_v1 <= Double.MIN_VALUE)) 
	    return 0;
	return error(_v1,_v2,cnt,_dvals,r)/cnt; 
    }

    double smallRatio(double r) {
	double rr = Math.min(r-SMALL_RATIO, SMALL_RATIO); 
	if (rr > Double.MIN_VALUE)
	    return rr;
	return rr/2;
    }
    public double error(double _v1, double _v2, int cnt, double _dvals[], double r) {
	// find ratio KL distance, blog(b/ar) + ar*log(ar/b) ==
	// b(log(b/ar) + ar/blog(ar/b)) = b*log(ar/b)*(ar/b-1) = b*log(rr)*(rr-1)
	double b =  _v2;
	double a =  _v1;

	if ((a <= Double.MIN_VALUE) && (b <= Double.MIN_VALUE)) 
	    // empty tuple..
	    return 0;
	if ((a <= Double.MIN_VALUE) && (r == LARGE_RATIO))
	    // aggregation of missing values.
	    return 0;
	if ((b <= Double.MIN_VALUE) && (r == 0))
	    // aggregation of missing values.
	    return 0;
	double logr = Math.log(r);
	if (r <= Double.MIN_VALUE)
	    logr = Math.log(Math.min(SMALL_RATIO, b/a));
	double logLr = 0;
	if (_dvals[2] > 0)
	    logLr = Math.log(Math.max(LARGE_RATIO, r+SMALL_RATIO));
	double logSr = 0;
	if (_dvals[3] > 0)
	    logSr = Math.log(smallRatio(r));
	double p = _dvals[0]  + _dvals[2]*logLr*logLr + _dvals[3]*logSr*logSr
	    - 2*logr*(_dvals[1] + _dvals[2]*logLr + _dvals[3]*logSr)
	    + logr*logr*cnt;
	return p; // sqrt(Math.max(p,0));
    }
};




