/*
RVals : Assumes some initial guesses of the ratios and these are refined as new tuples are read
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package diff;

import java.util.*;

import util.*;


    class RVals {
	double _rvals[];
	int _numR;
	double globalR;
	boolean externalRatios = false;

	public void init(double gr, int numR, float[] givenRatios) {
		globalR = gr;
		_numR = numR;
		_rvals = new double[_numR];

		if (givenRatios == null) {
		    // this initializationis needed to cover nodes before first
		    // call to setRvals.
		    int i;
		    _rvals[0] = globalR;
		    int sign = 1;
		    for (i = 1; i < _numR-1; i++) {
			_rvals[i] = globalR - sign*i*globalR/_numR;
			sign *= -1;
		    } 
		    if (_numR > 1) _rvals[i] = Tuple.LARGE_RATIO();
		} else {
		    externalRatios = true;
		    for (int i = 0; i < _numR; i++) {
			_rvals[i] = (double)givenRatios[i];
		    }
		}
	    }

	public boolean externalRatio() {return externalRatios;}
	public void setRVals(RVals rvals) {
	    if (externalRatios)
		return;
	    int rind = 0;
	    // reserve one slot for self, means keep the most redundant one
	    // at the last place.
	    double mindist = Double.MAX_VALUE;
	    int minrInd=0;
	    for (double r = rvals.val(rind); rind < rvals.size()-1; r = rvals.val(++rind)) {
		int r2ind = rind+1;
		for (double r2 = rvals.val(r2ind); r2 != -1; r2 = rvals.val(++r2ind)) {
		    double thisDist = Math.abs(r2-r);
		    if (thisDist >= mindist) continue;
		    minrInd = r2ind;
		    mindist = thisDist;
		}
	    }
	    rind = 0;
	    for (int i = 0; i < rvals.size(); i++) {
		if (i == minrInd) continue;
		_rvals[rind++] = rvals.val(i);
	    }
	    _rvals[_numR-1] = rvals.val(minrInd);
	    //	    System.out.println("Set R vals " + minrInd);
	    // rvals.output();
	    // output();
	}
	public int size() {return _numR;}
	public double val(int i) {if (i >= _numR) return -1; return _rvals[i];}

	public int closestRInd(double r) 
	{
	    // find closest r
	    int mInd = 0;
	    double minDiff = Double.MAX_VALUE;
	    for (int i = 0; i < _numR; i++) {
		double diff = Math.abs(r-_rvals[i]);
		if (diff < Double.MIN_VALUE)
		    return i;
		if (minDiff <= diff) continue;
		minDiff = diff;
		mInd = i;
	    }
	    //    assert(mInd < _numR);
	    return mInd;
	}
	void output() {
	    String st = "";
	    for (int i = 0; i < _numR; i++)
		st += ( _rvals[i] + " ");
	    System.out.println("Rvals " + st);
	} 
	public void update(double r, int rind) {
	    _rvals[rind] = r;
	}
	public  double getBestR(Tuple par, Tuple c1, Tuple c2) {
	    double v1 = par._v1 - (c1._v1 + c2._v1);
	    double v2 = par._v2 - (c1._v2 + c2._v2);
	    // 29 Aug 2K --- not clear why this was there in the first place..
	    // if ((v1 <= Double.MIN_VALUE) || (v2 <= Double.MIN_VALUE))
	    //	return Tuple.getRatio(c1._v1 + c2._v1, c1._v2 + c2._v2);
	    return Tuple.getRatio(v1,v2);
	}
    };
