package frontend;

/*
 * @(#)JDBCAdapter.java	1.9 98/08/26
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/**
 * An adaptor, transforming the JDBC interface to the TableModel interface.
 *
 * @version 1.20 09/25/97
 * @author Philip Milne
 */

import java.util.Vector;
import java.awt.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.*;
import javax.swing.event.TableModelEvent;

import java.awt.Dimension;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.Color;



public class CubeViewer extends MyTable {
    String colAttrName;
    String[]            columnNames;
    String[]            rowNames;
    Object[][]           data;
    int colStart = 0;
    int rowStart = 0;
    static Color IceBlue = new Color(204,204,255);
    static Color Wheat = new Color(255,218,181);
    static Color green = new Color(100,244,124);//my change	
    static int DefaultFontSize = 12;
    public static Font BoldFont = new Font("Dialog",Font.BOLD, DefaultFontSize);
    
    class CubeModel extends  AbstractTableModel {
	//////////////////////////////////////////////////////////////////////////
	//
	//             Implementation of the TableModel Interface
	//
	//////////////////////////////////////////////////////////////////////////

	// MetaData

	public String getColumnName(int column) {
	    if (column == 0)
		return "      ";
	    if ((columnNames == null) || (columnNames.length == 0))
		return "";
	    if (column == colStart)
		return colAttrName; // [column-numColAttrs];
	    return "";
	}

	public Class getColumnClass(int column) {
	    return String.class;
	    /*
	    if (column < colStart)
		return String.class;
	    return data[0][0].getClass();
	    */
	}

	public boolean isCellEditable(int row, int column) {
	    return false;
	}

	public int getColumnCount() {
	    if ((data == null) || (data.length == 0))
		return 0;
	    if ((columnNames == null) || (columnNames.length == 0))
		return colStart+1;
	    return columnNames.length + colStart;
	}

	// Data methods

	public int getRowCount() {
	    if ((data == null) || (data.length == 0))
		return 0;
	    if ((rowNames == null) || (rowNames.length == 0))
		return rowStart+1;
	    return rowNames.length + rowStart;
	}

	public Object getValueAt(int aRow, int aColumn) {
	    if ((aRow >= rowStart) && (aColumn >= colStart))
		return data[aRow-rowStart][aColumn-colStart];
	
	    if ((aColumn < colStart) && (aRow >= rowStart)) {
		if (rowNames.length > 0)
		    return rowNames[aRow-rowStart];
		else
		    return "";
	    }
	    if (aColumn >= colStart) {
		if (columnNames.length > 0)
		    return columnNames[aColumn-colStart];
		else
		    return "";
	    }
	    return rowAttrName;
	}

	public void setValueAt(Object value, int row, int column) {
	    ;
	}
	
    }
    CubeModel model;    
    DefaultTableCellRenderer labelRenderer;
    DefaultTableCellRenderer attrNameRenderer;
    DefaultTableCellRenderer highlightedCellRenderer;
    public CubeViewer() {
	super();
	model = new CubeModel();
	setModel(model);
    }
    
    class AttrName {
	String _name;
	AttrName(String val) {_name = val;}
	public String toString() {return _name;}
    }
    AttrName rowAttrName;

    public TableCellRenderer getCellRenderer(int row, int column) {
	if ((column < colStart) && (row < rowStart))
	    return attrNameRenderer;
	if ((column < colStart) || (row < rowStart))
	    return labelRenderer;
	if (highlighted(row,column))
	    return highlightedCellRenderer;
	return super.getCellRenderer(row,column);
	    
    }
    public void showResult(String colAttr, String rowAttr, String[] cols, String[] rows, Object[][] d) {
	colAttrName = colAttr;
	rowAttrName = new AttrName(rowAttr);
	columnNames = cols;
	rowNames = rows;
	data = d;
	rowStart = 1; // (cols.length == 0)?1:cols.length;
	colStart = 1; //(rows.length == 0)?1:rows.length;
	
	labelRenderer = new DefaultTableCellRenderer() {
		public void setValue(Object value) {
		    setBackground(Wheat);
		    if (value != null)
			setText(value.toString());
		}
	    };
	
	attrNameRenderer = new DefaultTableCellRenderer() {
		public void setValue(Object value) {
		    setBackground(IceBlue);
		    setFont(BoldFont);
//		    setTextColor(green);   //my change
		    if (value != null)
			setText(value.toString());
		      
		}
	    };


	highlightedCellRenderer = new DefaultTableCellRenderer() {
		public void setValue(Object value) {
		    setBorder(BorderFactory.createLineBorder(Color.red));
		    if (value != null)
			setText(value.toString());
		}
	    };

	int precision = 100;
	if ((data != null) && (data.length > 0) && (data[0][0] != null) && (data[0][0].getClass() == Double.class)) {
	    for (int i = 0; i < data.length; i++) {
		for (int j = 0; j < data[i].length; j++) {
		    if (data[i][j] == null)
			continue;
		    data[i][j] = Double.toString(Math.rint(((Double)data[i][j]).doubleValue()*precision)*1.0/precision);
		}
	    }
	}

	unselectAllCells();
	model.fireTableChanged(null); // Tell the listeners a new table has arrived.
	/*	for (int i = 0; i < model.getColumnCount(); i++)
	    for (int j = 0; j < model.getRowCount(); j++)
		System.out.println("" + i + " " + j + " " + model.getValueAt(j,i));
	*/
	/*
	setDefaultRenderer(getColumnClass(colStart), dataRenderer);
	for (int i = 0; i < colStart; i++) {
	    getColumn(getColumnName(i)).setCellRenderer(labelRenderer);
	    } */
	initColumnSizes(model);
	model.fireTableChanged(null); // Tell the listeners a new table has arrived.
	setCellSelectionEnabled(true);
	setRowSelectionAllowed(true);
	setColumnSelectionAllowed(true);

    }

    public void getSelectedDataCell(Vector attrName, Vector attrVal) {
	attrName.removeAllElements();
	attrVal.removeAllElements();
	int row = getSelectedRow();
	int col = getSelectedColumn();
	if ((row < rowStart) || (col < colStart))
	    return;
	attrName.addElement("Measure");
	attrVal.addElement(model.getValueAt(row,col));
	if (columnNames.length > col - colStart) {
	    attrName.addElement(colAttrName);
	    attrVal.addElement(columnNames[col-colStart]);
	}
	if (rowNames.length > row - rowStart) {
	    attrName.addElement(rowAttrName.toString());
	    attrVal.addElement(rowNames[row-rowStart]);
	}
    }

    private Vector highVals = new Vector();
    class TabCell {
	int row, col;
	TabCell(int r, int c) { row = r; col = c;}
	boolean equal(int r, int c) {return (row == r) && (col == c);}
    }
    private boolean highlighted(int row, int col) {
	for (int i = 0; i < highVals.size(); i++) {
	    TabCell cell = (TabCell) (highVals.elementAt(i));
	    if (cell.equal(row,col))
		return true;
	}
	return false;
    }
    public void unselectAllCells() {
	highVals.removeAllElements();
    }
    public void highlightSelectedCell() {
	int row = getSelectedRow();
	int col = getSelectedColumn();
	if ((row >= 0) && (col >= 0))
	    highVals.addElement(new TabCell(row,col));
	/*
	DataCell cell = (DataCell)getValueAt(row,cell);
	if (cell != null)
	    cell.setSelected();
	*/
    } 
}
