/*
ICubeOps : The 3 operators: DIFF, RELAX and INFORM are invoked from here and the results are displayed in a table 
    Copyright (C) 2001 Sunita Sarawagi, Gayatri Sathe

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package frontend;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.border.*;
import javax.swing.colorchooser.*;
import javax.swing.filechooser.*;
import javax.accessibility.*;
import javax.swing.table.*;
import javax.swing.tree.*;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import java.io.*;
import java.applet.*;
import java.net.*;

import cube.*;
import main.*;
import util.*;
import relax.*;

class ICubeOps {
    JPanel rootPanel;
    JPanel panel;
    ICubeFront cubeFront;
    JLabel progressLabel;
    JButton moreDiffButton = null;
    
    Vector attrNames;
    Vector attrVals;
    Vector attrNames1;
    Vector attrNames2;
    Vector attrVals1; 
    Vector attrVals2; 
    MTerm firstTerm,secondTerm;
    Term diffAggLevel = null;
   
    boolean selectFirst = false;
    boolean diffPending = false;
    boolean relaxPending = false;

    class DiffResultTableModel extends ResultTableModel{
	TableSorter sorter;
	Result res;
	DiffResultTableModel() {;}
	DiffResultTableModel(Result r) {init(r);}
	void init(Result r) {res = r; fireTableChanged(null);}
	public Term getTermAt(int row) {
	    if (sorter != null) {
		row = sorter.getMappedRow(row);
	    }
	    return res.fullTuple(row);
	}
	public int getColumnCount() { 
	    if ((res != null) && (res.numTups() > 0))
		return res.tup(0).numAttrs()+4; 
	    return 0;
	}
	public int getRowCount() { 
	    if ((res != null) && (res.numTups() > 0))
		return res.numTups();
	    return 0;
	}
	public Object getValueAt(int row, int col) {
	    int numAttrs = res.tup(row).numAttrs();
	    if (col < numAttrs)
		return res.tup(row).getString(col);
	    if (col == numAttrs)
		return new Float(res.tup(row).meas1());
	    if (col == numAttrs+1)
		return new Float(res.tup(row).meas2());
	    if (col == numAttrs+2)
		return new Float(res.tup(row).ratio());
	    return new Float(res.error(row));
	}
	public String getColumnName(int col) {
	    int row = 0;
	    int numAttrs = res.tup(row).numAttrs();
	    if (col < numAttrs)
		return res.tup(row).desc(col).name();
	    if (col == numAttrs)
		return res.firstMeasureName();
	    if (col == numAttrs+1)
		return res.secondMeasureName();
	    if (col == numAttrs+2)
		return "ratio";
	    return "Error";
	}
	public Class getColumnClass(int c) {
	    return getValueAt(0, c).getClass();
	}
	public boolean isCellEditable(int row, int col) {return false;}
	public void setValueAt(Object aValue, int row, int column) { return; }
    };

    class RelaxResultTableModel extends DiffResultTableModel {
	RelaxResult relRes;
	String path=new String();
	RelaxResultTableModel(Result r) {super(r); init(r);}
	void init(Result r) {
	    relRes = (RelaxResult)r; 
	    fireTableChanged(null);
	}
	public Term getTermAt(int row) {
	    return relRes.fullTuple(row);
	}
	public int getColumnCount() { 
	    if ((relRes != null) && (relRes.numTups() > 0))
		return relRes.tup(0).numAttrs() + 4;   
	    return 0;
	}
	public int getRowCount() { 
	    if ((relRes != null) && (relRes.numTups() > 0))
		return relRes.numTups();
	    return 0;
	}
	public Object getValueAt(int row, int col) {
	    int numAttrs = relRes.tup(row).numAttrs();
	    if (col == 0) {
		if (row == 0)
		    return "Problem case" ;
		if (row==1) {
		    path="1";
		    return "G"+path;
		}
		else{	
		   int curr_level = getLevel(row);
		   int prev_level = getLevel(row-1);
		   if (curr_level == prev_level) 
			path = incrPath(path);   // increment the last level of path
		   else if (curr_level > prev_level)
			path+=".1";
		   else {    // extract path of the previous row at the same level and incr the last number
		        int diff=prev_level-curr_level;
		        while(diff>0) {
			   path = path.substring(0,path.lastIndexOf("."));
			   diff--;
		        }
			path = incrPath(path);  // increment the last level of path
		   }
		   if (relRes.isException(row))
		      return appendString(" ",3*curr_level)+"E"+path;	
		   else return "G"+path;	
		}
	    }
	    if (col == getColumnCount()-3) 
		return displayPrecision(new Float(relRes.tup(row).meas1()));
//		return new Float(relRes.tup(row).meas1());
	    if (col == getColumnCount()-2) 
		return displayPrecision(new Float(relRes.tup(row).meas2()));
//		return new Float(relRes.tup(row).meas2());
	    if (col == getColumnCount()-1) 
 	        return new Integer(relRes.count(row));
 
	    col--;
	    Term thisTup = relRes.tup(row);

	// col is blank only when (thiscol == probcol) and parent(thiscol) != *
	    if (row>0)
	        if ((thisTup.getString(col).equals(relRes.tup(0).getString(col))) &&  (!relRes.tup(assignParent(row)).getString(col).equals("*")))
		return "";

	    return thisTup.getString(col);
	}
	public String getColumnName(int col) {
	    int row = 0;
	    int numAttrs = relRes.tup(row).numAttrs();
	    if (col == 0)
		return "Type";	
	    if (col == getColumnCount()-3)
		return relRes.firstMeasureName();
	    if (col == getColumnCount()-2)
		return relRes.secondMeasureName();
	    if (col == getColumnCount()-1)
   	        return "Count";
	    col--;
	    return relRes.tup(row).desc(col).name();
	}
	public Class getColumnClass(int c) {
	    return getValueAt(0, c).getClass();
	}
	public boolean isCellEditable(int row, int col) {return false;}
	public void setValueAt(Object aValue, int row, int column) { return; }
        public int getLevel(int curr){
	    int count = 0;
	    int parent=curr;
	    while (parent!=0){
	        parent = assignParent(parent);
	        count++;
	    }
 	    return count;  
        }
	public String incrPath(String s){
	    int d=s.lastIndexOf(".");
	    String lastchar = s.substring(d+1);
	//    System.out.println("lastchar = "+lastchar);
	    int i=Integer.parseInt(lastchar);
	    lastchar = Integer.toString(i+1);
	    s=s.substring(0,d+1) + lastchar;
	    return s;		
	}
        public int assignParent(int e){
	    Term curr = relRes.tup(e);
	    for(int i=e-1; i>0; i--){
	       Term prev = relRes.tup(i);
	       for(int col=0; col<=relRes.tup(e).numAttrs(); col++){
		  if (col==relRes.tup(e).numAttrs()) return(i);
		  if (!curr.attrVal(col).same(prev.attrVal(col))) {
		 	if (prev.constant(col)) break;
		 	if (curr.range(col)) break;
		  }
	       }	
	    }
	    return(0);
        }	 
        public String appendString(String c, int n){
	    String s = new String("");
	    for (int i=0; i<n; i++)
	        s+=c;
	    return s;
        }
	public Float displayPrecision(Float f){
	    String fs = f.toString();
	    int i = fs.indexOf(".");
	    return Float.valueOf(fs.substring(0,i+2));   
	}
    };
    ResultTableModel model = null;
    MyTable resTable = null;
    Result currResult;
    
    void makeNewTable() {
	if (resTable == null) {
	    resTable = new MyTable();
	    resTable.setCellSelectionEnabled(false);
	    resTable.setColumnSelectionAllowed(false);
	    JPanel tpanel = new JPanel();
	    tpanel.setLayout(new BoxLayout(tpanel, BoxLayout.Y_AXIS));
	    tpanel.add(new JScrollPane(resTable));//, BorderLayout.CENTER);
	    tpanel.add(newDiffButton());//, BorderLayout.SOUTH);
	    panel.add(tpanel, BorderLayout.CENTER);
	} 
    }
    void setTableModel(AbstractTableModel m, ResultTableModel m1) {
	model = m1;
	moreDiffButton.setEnabled(false);
	resTable.setModel(m);
	resTable.initColumnSizes(m);
	m.fireTableChanged(null);
    }

    void showResult(Result res) {
	currResult = res;
	DiffResultTableModel model1 = new DiffResultTableModel(res);
	model1.sorter = new TableSorter(model1);
	model1.sorter.addMouseListenerToHeaderInTable(resTable);
	setTableModel(model1.sorter, model1);
	if (res != null) {
	    moreDiffButton.setEnabled(true);
	}
    }
    	
    void showRelaxResult(Result res) {   
	currResult = res;
	ResultTableModel model1 = new RelaxResultTableModel(res);
	setTableModel(model1, model1);
    } 

    Result getResult() {return currResult;}

    Term getSelectedTerm() {
       	if ((resTable == null) || (model == null))
	    return null;
	int row = resTable.getSelectedRow();
	if (row >= 0)
	    return model.getTermAt(row);
	return null;
    }

    void putMessage(String txt) {
	progressLabel.setText(txt);
    }

    void getDiff() {
	diffPending = false;
	try {
	    putMessage("Getting difference..");
	    Result res = server().getDiff(firstTerm, secondTerm,diffAggLevel);
	    showResult(res);
	    putMessage("Done");
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    void relax() {
	relaxPending = false;
	try {
	    putMessage("Relaxing around the tuple...");
	    Result res = server().relax(firstTerm, secondTerm);
	    showRelaxResult(res);
	    if (res == null)
		putMessage("No generalization possible");
	    else
		putMessage("Done");
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }
		    
    void selectTwoCells() {
	progressLabel.setText("Select first cell.");
	selectFirst = true;
    }

    void cellSelected() {
	if (selectFirst) {
	    attrNames1 = attrNames;
	    attrVals1= attrVals;
	    progressLabel.setText("Select  second cell.");	    
	    selectFirst = false;
	} else {
	    attrNames2 = attrNames;
	    attrVals2 = attrVals;
	    if (diffPending) {
		getDiff();
	    } else if (relaxPending) {
		relax();
	    }
	}
    }

    void cellSelected(MTerm t) {
	if (selectFirst) {
	    firstTerm = t;
	    progressLabel.setText("Select  second cell.");	    
	    selectFirst = false;
	} else {
	    secondTerm = t;
	    if (diffPending) {
		getDiff();
	    } else if (relaxPending) {
		relax();
	    }
	}
    }

    JButton newDiffButton() {
	moreDiffButton = new JButton("More..");
	moreDiffButton.setEnabled(false);
	moreDiffButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    try {
			if (getResult() == null)
			    return;
			putMessage("Getting more results..");
			showResult(server().getDiff(getResult()));
			putMessage("Done");
		    } catch (Exception exp) {
			exp.printStackTrace();
		    }
		}
	    });
	return moreDiffButton;
    }
     
    JMenu addMenus(JToolBar toolBar, JToolBar tableToolBar, JMenuBar menuBar) {
	JMenuItem mi;

	JButton p = ICubeFront.addTool(toolBar, "Select",true,null);
	p.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    MTerm t = cubeState().getSelectedDataCell();
		    if (t != null) {
			cellSelected(t);
		    }
		    cubeState().getViewer().highlightSelectedCell();
		}
	    });

	
	JButton b = ICubeFront.addTool(toolBar, "Diff",true,null);
	b.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    cubeState().getViewer().unselectAllCells();
		    diffPending = true;
		    selectTwoCells();
		}
	    });
	
	p = ICubeFront.addTool(toolBar, "Relax",true,null);
	p.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    cubeState().getViewer().unselectAllCells();
		    relaxPending = true;
		    selectTwoCells();
		}
	    });


	p = ICubeFront.addTool(tableToolBar, "Context",true,null);
	p.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    Term selTerm = getSelectedTerm();
		    if (selTerm != null) {
			cubeState().setContext(selTerm);
		    }
		}
	    });

	JMenu menu = (JMenu) menuBar.add(new JMenu("Options")); 
	JMenuItem mitem = (JMenuItem) menu.add(new JMenuItem("Diff parameters"));
        mitem.setMnemonic('D');
	mitem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    String pars = JOptionPane.showInputDialog(panel, "Semi-colon seperated parameters for Diff:");
		    try {
			server().setDiffParams(pars);
		    } catch (Exception exp) {
			exp.printStackTrace();
		    }
		}
	    });

	mitem = (JMenuItem) menu.add(new JMenuItem("Relax parameters"));
        mitem.setMnemonic('R');
	mitem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    String pars = JOptionPane.showInputDialog(panel, "Semi-colon seperated parameters for Relax:");
		    try {
			server().setRelaxParams(pars);
		    } catch (Exception exp) {
			exp.printStackTrace();
		    }
		}
	    });
	
	mitem = (JMenuItem) menu.add(new JMenuItem("Diff drill down level"));
        mitem.setMnemonic('A');
	mitem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    String dimIndices = JOptionPane.showInputDialog(panel, "Semi-colon seperated attribute indices");
		    diffAggLevel = cubeState().setSpecifiedAttributesToRange(dimIndices);
		}
	    });
	
	return menu;
    }
    
    ICubeState cubeState() {return cubeFront.getCubeState();}
    ICubeServer server() {return cubeFront.getCubeServer();}
    
    void close() { 
	selectFirst = false;
	diffPending = false;
	relaxPending = false;
    }
    
    ICubeOps(JPanel rPanel, JPanel p, JToolBar toolBar, JMenuBar menuBar, ICubeFront cfront) { 
	rootPanel = rPanel;
	panel = p;
	panel.setLayout(new BorderLayout());
	progressLabel = new JLabel("Ready..");

	JToolBar tableToolBar = new JToolBar();
	JPanel tablePanel = new JPanel();
	tablePanel.setLayout(new BorderLayout());
	tablePanel.add(tableToolBar, BorderLayout.SOUTH);
	panel.add(tablePanel, BorderLayout.NORTH);

	panel.add(progressLabel, BorderLayout.SOUTH);
	cubeFront = cfront;
	
	JMenu menu = addMenus(toolBar, tableToolBar, menuBar);
	makeNewTable();


	/*------------------------------------------------
	 * Maxent not provided in this release. Sept 2001
	 */
	/*	MaxEntFront inform = new MaxEntFront();
	inform.init(toolBar, tableToolBar, tablePanel,menu,this);
	*/
    }
}

