/*
ICubeState : Maintains the state of the data cube as viewed by the user
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package frontend;
import cube.*;
import main.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.border.*;
import javax.swing.colorchooser.*;
import javax.swing.filechooser.*;
import javax.accessibility.*;
import javax.swing.table.*;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import java.io.*;
import java.applet.*;
import java.net.*;

import util.*;

class ICubeState {
    static final char PAGE = 'P';
    static final char ROW = 'R';
    static final char COLUMN = 'C';
    static final int ROWINDEX = 1;
    static final int COLUMNINDEX = 2;

    ICubeServer server;
    Cube cube;
    Term selectedVals;
    char orient[];
    int order[];
    CubeViewer cubeViewer;

    
    JComboBox comboBox[];
    boolean changeState = true;

    ICubeState(ICubeServer s, Cube c, CubeViewer gridView) {
	server = s;
	cubeViewer  = gridView;
	cube = c;
	
	selectedVals = new Term(MdUtils.getColDesc(cube));
	orient = new char[cube.numAttrs()];
	for (int i = 0; i < orient.length; i++) {
	    orient[i] = PAGE; 
	}
	order = new int[cube.numAttrs()];

	comboBox = new JComboBox[cube.numAttrs()];
    }
    void setMembers(JComboBox members, int attrNum, int dim, int lvl) {
	comboBox[attrNum] = members;
	members.addItem("(All)");
	if (cube.numMembers(dim,lvl) > ICubeFront.MAX_MEMBERS) {
	    members.addItem("(Too many Items)");
	    return;
	}
	members.addItem("(Row)");
	members.addItem("(Column)");
	for (int m = 0; m < cube.numMembers(dim,lvl); m++) {
	    members.addItem(cube.memberName(dim,lvl,m));
	} 
	
    }
    private boolean isRow(int index) {return (index == ROWINDEX);} 
    private boolean isColumn(int index) {return (index == COLUMNINDEX);} 
    private char getOrientation(int index) {
	if (isRow(index))
	     return ROW;
	else if (isColumn(index))
	    return COLUMN;
	else 
	    return PAGE;
    }

    private Object getVal(int dim, int lvl, int index, String val) {
	// create a term structure 
	if (index == 0) 
	    return Term.AGGREGATED;
	else if (isRow(index) || isColumn(index))
	    return Term.RANGE;
	else {
	    return val;
	}
    }

    void setAttribute(int attrNum, char thisOrient, Object val, String name) {
	selectedVals.setVal(val, attrNum);
	switch (thisOrient) {
	case PAGE: 
	    if (!selectedVals.aggregated(val)) {
		// TODO what if the object pointer is not the same?
		comboBox[attrNum].setSelectedItem(name);
	    } else
		comboBox[attrNum].setSelectedIndex(0);
	    break;
	case ROW:
	    comboBox[attrNum].setSelectedIndex(ROWINDEX);
	    break;
	case COLUMN:
	    comboBox[attrNum].setSelectedIndex(COLUMNINDEX);
	    break;

	}
	orient[attrNum] = thisOrient;
    }
    
    void registerChange(int attrNum, int dim, int lvl, int index) {
	boolean change = false;
	
	// set the orientation..
	char thisOrient = getOrientation(index);
	Object thisVal = getVal(dim,lvl, index, (String)comboBox[attrNum].getItemAt(index));

	if (thisOrient != orient[attrNum]) {
	    change = true;
	    if (thisOrient != PAGE) {
		// make sure that no other attribute has the same orientation if row/col
		for (int i = 0; i < orient.length; i++) {
		    if (orient[i] != thisOrient)
			continue;
		    orient[i] = PAGE;
		    comboBox[i].setSelectedIndex(0);
		}
	    }
	    orient[attrNum] = thisOrient;	    
	}
	
	if (thisVal != selectedVals.val(attrNum)) {
	    change = true;
	    selectedVals.setVal(thisVal, attrNum);
	}
	
	if (change == false)
	    return;
	if (!changeState)
	    return;
	fetchData();
    }

    void fetchData() {
	// fix the order based on row and columns..
	int orderIndex = 0;
	for (int i = 0; i < orient.length; i++) {
	    if (orient[i] == ROW) 
		order[orderIndex++] = i;
	}
	for (int i = 0; i < orient.length; i++) {
	    if (orient[i] == COLUMN) 
		order[orderIndex++] = i;
	}
	
	try {
	    // ask for data.
	    showResult();
	    // for (int i = 0; i < cubeViewListeners.size(); i++) {
	    // ((CubeViewListener)cubeViewListeners.elementAt(i)).notify();
	    // }
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.err.println("Caught ArrayIndexOutOfBoundsException: " +
			       e.getMessage());
	    e.printStackTrace();
	}  catch (Exception e) {
	    e.printStackTrace();
	} 
    }

    int numAttrs() {return selectedVals.numAttrs();}
    public void showResult() {
	try {
	    int axis1 = -1, axis2 = -1;
	    for (int i = 0; i < numAttrs(); i++) {
		if (orient[i] == ROW) 
		    axis1 = i;
		if (orient[i] == COLUMN)
		    axis2 = i;
	    }
	    CubeData res = server.subCube(selectedVals, axis1, axis2); 
	    String[] columnNames = res.axis2();
	    String[] rowNames = res.axis1();
	    
	    String rowAttrName = null;
	    String colAttrName = null;
	    for (int i = 0;i < numAttrs(); i++) {
		if (orient[i] == ROW)
		    rowAttrName = selectedVals.desc(i).name();
		if (orient[i] == COLUMN)
		    colAttrName = selectedVals.desc(i).name();
	    }
	    cubeViewer.showResult(colAttrName,rowAttrName,
				  columnNames, rowNames, res.data());
	}
        catch (Exception ex) {
			ex.printStackTrace();
            System.err.println(ex);
        }
    }
    
    public boolean getSelectedDataCell(Vector attrName, Vector attrVal) {
	cubeViewer.getSelectedDataCell(attrName, attrVal);
	if (attrName.size() == 0)
	    return false;
	for (int i = 0; i < numAttrs(); i++) {
	    if ((orient[i] == PAGE) && selectedVals.constant(i)) {
		attrName.addElement(selectedVals.desc(i).name());
		attrVal.addElement(comboBox[i].getSelectedItem());
	    }
	}
	return true;
    }
  
    public MTerm getSelectedDataCell() {
	Vector attrName = new Vector();
	Vector attrVal = new Vector();
	cubeViewer.getSelectedDataCell(attrName, attrVal);
	MTerm t = new MTerm(selectedVals);
	if (attrName.size() == 0)
	    return null;
	t.setMeasure(Float.valueOf(attrVal.elementAt(0).toString()).floatValue());
	for (int i = 0; i < numAttrs(); i++) {
	    if (orient[i] == PAGE) 
		continue;
	    boolean valSelected = false;
	    for (int j = 1; j < attrName.size(); j++) {
		if (t.desc(i).name().equals(attrName.elementAt(j))) {
		    t.setVal(attrVal.elementAt(j), i);
		    valSelected = true;
		    break;
		}
	    }
	    if (!valSelected) return null;
	}
	return t;
    }
    public Term getContext() {
	return selectedVals;
    }
    public void setContext(Term t) {
	boolean rowSet = false;
	boolean columnSet = false;
	changeState = false;
	for (int i = t.numAttrs()-1; i >= 0; i--) {
	    char thisOrient = PAGE;
	    if (t.range(i)) {
		if (!columnSet) {
		    columnSet = true;
		    thisOrient = COLUMN;
		} else if (!rowSet) {
		    rowSet = true;
		    thisOrient = ROW;
		} else
		    System.out.println("Cannot display more than two dimensions");
	    }
	    setAttribute(i, thisOrient, t.val(i), t.getString(i));
	}
	changeState = true;
	fetchData();
    }
    CubeViewer getViewer() {return cubeViewer;}
    
    Term setSpecifiedAttributesToRange(String attrList) {
	if ((attrList == null) || (attrList.length() == 0))
	    return null;
	Term t = new Term(selectedVals.termDescr());
	// for (int i = 0; i < t.numAttrs(); t.setVal(Term.AGGREGATE, i++));
	StringTokenizer nameTok = new StringTokenizer(attrList, ";");
	for (int i = 0; nameTok.hasMoreTokens(); i++) {
	   int indx = Integer.parseInt(nameTok.nextToken());
	   if ((indx < t.numAttrs()) && (indx >= 0))
	       t.setVal(Term.RANGE, indx);
	}
	return t;
    }
}


