/*
ICubeServerC : Initializing the ICube server. 
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package main;
//import maxent.*;
import diff.*;
import relax.*;
import cube.*;
import util.*;

import java.util.*;
import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;

public class ICubeServerC 
    extends UnicastRemoteObject
    implements ICubeServer
{
    private Cube cube;
    private String usrName;
    private Diff diff;
    private Relax relax;
    private String maxParams;
    private String diffParams;
    private String relaxParams;
    TermDesc coldesc;

    public ICubeServerC() throws RemoteException {
	super();
    }

    private void init() {
	//	maxent = null;
	diff = null;
	relax = null;
	maxParams = null;
	diffParams = null;
	relaxParams = null;
    }

    public Cube newCube(String cubeName, String usr, String passwd) throws RemoteException, Exception {
	return newCube(cubeName, usr, passwd, "RelOLAP");
    }
    public Cube newCube(String cubeName, String usr, String passwd, String provider) throws RemoteException, Exception {
	try {
	usrName = usr;
	cube = MdUtils.newCube(provider, cubeName, usr, passwd, true);
	coldesc = MdUtils.getColDesc(cube);
	init();
	return cube;
	} catch (Exception e) {
	    e.printStackTrace();
	    throw new Exception(e.getMessage());
	}
     }
    private boolean rolap() {
	return RelCube.class.isAssignableFrom(cube.getClass());
    }

    private void checkInitDiff()  throws Exception
    {
	if (diff == null) {
	    diff = new Diff(cube, new DiffParams(diffParams));
	}
    }
    private void checkInitRelax()  throws Exception
    {
        if (relax == null) {
	    relax = new Relax(cube, new RelaxParams(relaxParams));
	}
    }


    public CubeData subCube(Term t, int axis1, int axis2) throws RemoteException, Exception {
	return cube.subCubeExt(t, axis1, axis2);
    }

    public Result getDiff(Vector attrNames1, Vector attrVals1, Vector attrNames2, Vector attrVals2) throws RemoteException, Exception {
	return diff.getDiff(attrNames1, attrVals1, attrNames2, attrVals2);
    }

    public Result getDiff(Term firstTerm, Term secondTerm, Term aggLevel) throws RemoteException, Exception {
	if (diff == null) {
	    diff = new Diff(cube, new DiffParams(diffParams));
	}
	return diff.getDiff( firstTerm,  secondTerm, aggLevel);
    }
    public Result getDiff(Result res) throws RemoteException, Exception {
	if (diff == null) {
	    diff = new Diff(cube, new DiffParams(diffParams));
	}
	return diff.getDiff(res);
    }
    public void setDiffParams(String opts) throws RemoteException {
	diffParams = opts;
	if (diff != null) {
	    diff.setParams(opts);
	} 
    }

    public void setRelaxParams(String opts) throws RemoteException {
	relaxParams = opts;
	if (relax != null) {
	    relax.setParams(new RelaxParams(opts));
	} 
    }
    public Result relax(Vector attrNames1, Vector attrVals1, Vector attrNames2, Vector attrVals2) throws RemoteException, Exception {
	MTerm t1 = MdUtils.getMTermMapping(cube,coldesc,attrNames1, attrVals1);
	MTerm t2 = MdUtils.getMTermMapping(cube,coldesc,attrNames2,attrVals2);
	return relax(t1,t2);
    }
    public Result relax(MTerm t1, MTerm t2) throws RemoteException, Exception
    {
	System.out.println(t1.toFullString());
	System.out.println(t2.toFullString());
	checkInitRelax();
	return relax.relax(t1, t2);
    }

    /*------------------------------------------------
     * Maxent not provided in this release. Sept 2001
     */
    /*
    private MaxEntM maxent;
    private void checkInitMaxent() throws Exception
    { 
	if (maxent == null) {
	    if (rolap())
		maxent = new MaxEntM((RelCube)cube, usrName, new AlgParams(maxParams));
	    else throw new Exception("Not implemented");
	}
    }
    public void setMaxentParams(String opts) throws RemoteException {
	maxParams = opts;
	maxent = null;
	try {
	    checkInitMaxent();
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }
    public double markSeen(Vector attrNames, Vector attrVals) throws RemoteException, Exception 
    {
	checkInitMaxent();
	return maxent.queueConstraint(attrNames, attrVals);
    }
    public double markSeen(Term t) throws RemoteException, Exception 
    {
	checkInitMaxent();
	return maxent.queueConstraint(t);
    }
    public Result topNSurprises() throws RemoteException, Exception 
    {
	checkInitMaxent();
	return maxent.returnSurprises();
    }
    public float getInformationContentPercent() throws RemoteException, Exception 
    {
	checkInitMaxent();
	return maxent.findInformation()*100;
    }
    public Term informativeNeibr(Term t) throws RemoteException, Exception 
    {
	checkInitMaxent();
	return maxent.informativeNeibr(t);
    }
    public InfoResult memberInfoQuery(Term dimLvl, Term memLvl, boolean sortOrder, int k) 
	throws RemoteException, Exception
    {
	checkInitMaxent();
	return maxent.memberInfoQuery(dimLvl, memLvl, sortOrder, k);
    }
    */
    
    public static void main(String args[])
    {
	// Create and install a security manager
	System.setSecurityManager(new RMISecurityManager());
	
	try {
	    ICubeServerC obj = new ICubeServerC();
	    Naming.rebind("iCubeServer", obj);
	    System.out.println("iCubeServer bound in registry");
	} catch (Exception e) {
	    System.out.println("iCubeServer err: " + e.getMessage());
	    e.printStackTrace();
	}
    }
}
