/*
Relax : Takes as input 2 Terms and returns the generalizations along with exceptions to each.
    Copyright (C) 2001  Gayatri Sathe, Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package relax;

import java.util.*;
import cube.*;
import util.*;
import diff.*;

class RelaxStatistics {
        float stats[] = null;    // 1:total 2:Rval 3:Sval
        float stat(int i) {return stats[i];}
	int count() {return (int)stats[0];}
        RelaxStatistics(float s[]) {
            stats = s;
        }
	float Rval(){ return stats[1];}
	float Sval(){ return stats[2];}
        void display() {
            for (int i=0; i<stats.length; i++)
                System.out.print("i:"+stats[i]+" ");
            System.out.println();
        }
	boolean relaxPossible(){
		return (Sval()>Rval());
	}
	public void getStatistics(Term nterm1, Term nterm2, Cube cube, Relation rel) throws Exception{
            float[] counts = new float[3];
            TermDesc tdesc = MdUtils.getTermDescForRange(nterm1);
            // getting the merged subcube of two terms
            QueryHandle q = cube.concatCubes(nterm1,nterm2,tdesc,
                        new MeasureExpr("sum("+cube.measure(0)+")","v0"),
                        new MeasureExpr("sum("+cube.measure(0)+")","v1"),
                        null);
            MeasureExpr meas[] = new MeasureExpr[3];
            meas[0] = new MeasureExpr("count(*)","total");
	    meas[1] = new MeasureExpr("sum("+rel.Rfunc("v0","v1")+")","R");
	    meas[2] = new MeasureExpr("sum("+rel.Sfunc("v0","v1")+")","S");

            Term t = new Term(nterm1);
            for (int i=0; i<t.numAttrs(); i++)
             	t.setVal(Term.AGGREGATED, i);
            QueryHandle qh = cube.subCube(t,null,meas,q,null);

            CubeRows c = cube.executeQuery(qh);
	    while (c.next()) {
		counts[0] = (float)c.getDouble(1);
	 	counts[1] = (float)c.getDouble(2);
		counts[2] = (float)c.getDouble(3);	
	    }
	    c.close();
	    stats = rel.getRS(counts);
    	}
};


public class Relax {
    RelaxParams pars;
    Cube cube;
    BitSetPlus tmp = new BitSetPlus(); // for temporary use..
    int state[];  // minimum cost partition
    double mincost;
    	 
    public Relax(Cube c,RelaxParams params) {
    	cube = c;
	pars = params;
    }
    public void setParams(RelaxParams params) {
	pars = params;
    }

    public Result relax(MTerm t1, MTerm t2) throws Exception {
	t1.validate(cube);
	t2.validate(cube);
	// needed for booleanRelation
	pars.setMeasures(t1.measure(),t2.measure());
	return relax(t1,t2,pars.getRel());
    }

    //
    // SS-- give all timing measurements as an option in RelaxParams.
    //   --- use the debug flag to limit output
    public Result relax(MTerm t1, MTerm t2, Relation rel) throws Exception {
	long start=0,end=0;
	recordTime time = new recordTime(pars);
	start = time.current(); 
	//check if generalization is possible along every dimension
	BitSetPlus prevBitSet = null;
	if (t1.isEqual(t2,null))
	    return null;
	Vector L[] = new Vector[t1.numAttrs()+1];
	L[0] = new Vector();//store bitvectors of gen 

	long statstart=0, statend=0, genstart=0, genend=0, gentime=0;
	long timestat=0;
	long dbgentime=0;
        int num_tuples=0;
	genstart = time.current(); 
	for (int i = t1.numAttrs()-1; (i >= 0); i--) {
	    BitSetPlus mask = new BitSetPlus(t1.numAttrs());//set bits when gen possible
	    if (! t1.attrVal(i).same(t2.attrVal(i))) //gen along dim which are equal
		continue;
	    if (t1.aggregated(i))
		continue;
	    int j = i+1; 
	    	    
	    if ((j < t1.numAttrs()) && (t1.desc(j).dim() == t1.desc(i).dim()) && t1.constant(j)) {
		if (prevBitSet != null) {
		    mask.or(prevBitSet);
		} else
		    continue;
	    }
	    mask.set(i); 
//	    System.out.println("mask :"+mask);
	    statstart=time.current(); 	
	    RelaxStatistics relstat = findMeritOfGen(t1, t2, mask, rel);
	    statend = time.current(); 
	    num_tuples+=relstat.count();
            gentime=statend-statstart;		
	    timestat+=gentime;
//	    System.out.println("----DB Time for gen "+mask+ " = "+gentime);
	    dbgentime+=gentime;
	    if (relstat.relaxPossible()) {
		prevBitSet = mask;
		int k = mask.numOnes()-1;
		if (L[k] == null) {
	          L[k] = new Vector();
		}
		L[k].addElement(mask);
	    } else	
		prevBitSet = null;
	}
	if (L[0].isEmpty()) {
	    System.out.println("no generalization possible");
	    return null;
	}
	System.out.println("single dim gen vector");
	printBitList(L[0]);	
	//
	// now check for generalizations along higher dimensions..
	// a higher dimensional generalization is checked only 
	// if lower dimensional ones are possible.
	//
	for (int k = 0; (L[k] != null) && ((Vector)L[k]).size() > 0; k++) { 
	    Vector Lk = (Vector)L[k];
	    Vector candGens = getCandidates(Lk);
	    if (candGens.size() == 0)
		break;
	    if (L[k+1] == null) 
		L[k+1] = new Vector();
	    for (int i = 0; i < candGens.size(); i++) {
		BitSetPlus v = (BitSetPlus)candGens.elementAt(i);
	        statstart=time.current(); 	
		RelaxStatistics relstat = findMeritOfGen(t1, t2, v, rel);
	        statend=time.current(); 
	        num_tuples+=relstat.count();
                gentime=statend-statstart;		
	        timestat+=gentime;
//	        System.out.println("----DB Time for gen "+v+ " = "+gentime);
		dbgentime+=gentime;
		if (relstat.relaxPossible()) {
		    L[k+1].addElement(v);
		}
	    }
	    System.out.print("candidates");
	    printBitList(candGens);
	    System.out.print("next level candidates");
	    printBitList(L[k+1]);
	}    
//        time.printStats("Statistics collection time",timestat);
//	time.printStats("Time spent in DB for generalization",dbgentime);
//	time.printStats("Tuples covered",num_tuples);
	//
	// Remove all subsumed bitmaps..
	//
	Vector finalVecs = new Vector();
	for (int k = 0; (k < L.length) && (L[k] != null); k++) {
	    Vector Lk = (Vector)L[k];
	    for (int i = 0; i < Lk.size(); i++) {
		BitSetPlus v = (BitSetPlus)Lk.elementAt(i);
		boolean subsumed = false;
		// check if this bitset is subsumed by the level above it
		if ((k+1 < L.length) && (L[k+1] != null)) {
		    for (int j = 0; j < L[k+1].size(); j++) {
			if (((BitSetPlus)L[k+1].elementAt(j)).subsumes(v)) {
			    subsumed = true;
			    break;
			}
		    }
		}
		if (!subsumed) {
		    finalVecs.addElement(v);
		}
	    }
	}
	genend = time.current(); 
// 	time.printStats("Time for generalization",genend-genstart);
	System.out.println("final vector after multidim gen");
	for (int j = 0; j < finalVecs.size(); j++)
	    System.out.print(finalVecs.elementAt(j)+ " ");
	System.out.println(); 	

	//
	// The list of maximal generalizations is obtained.
	// Now get the exceptions for each possible generalization..
	ETerm tup = null;
	int G = finalVecs.size();
	AllDiffResult r[] = new AllDiffResult[G];
	long startapp,endapp;
	startapp = time.current(); 
	for (int i = 0; i < G; i++) {
	    BitSetPlus v = (BitSetPlus)finalVecs.elementAt(i);
	   System.out.println("getting exceptions");
	   r[i] = getSummarizedExceptions(t1,t2,v,rel);
	}
	endapp = time.current(); 
	//time.printStats("Time for getting exceptions",endapp-startapp);

	//
	// Now need to partition the user-specified set of N slots amongst 
	// the G slots to get minimum total error
	state = getBestPartition(G,pars.resultSize(),r);
	for(int i=0; i<G; i++)
	 System.out.print(state[i]+" ");
	System.out.println();

	RelaxResult res = new RelaxResult(t1, t2, null, cube);
	for(int i=0; i<G; i++) {
	   DiffResult dr = r[i].getResult(state[i]); //get result of specified size for ith generalization
	   BitSetPlus v = (BitSetPlus)finalVecs.elementAt(i); //get mask for ith generalization
	   res.addGeneralization(v,dr);
	}
	res.output();
	end = time.current();
//	time.printStats("Total time taken",end-start);
	time.printStatistics(timestat,dbgentime,num_tuples,genend-genstart,endapp-startapp,end-start);
	return res;
    }

    Term nterm1, nterm2;
    RelaxStatistics findMeritOfGen(Term t1, Term t2, BitSetPlus mask, Relation rel) throws Exception 
    {
	prepareNTerms(t1, t2, mask);
//	System.out.println(nterm1);
//	System.out.println(nterm2);
	RelaxStatistics relstat = new RelaxStatistics(null);
	relstat.getStatistics(nterm1,nterm2,cube,rel);
	mask.setCount((int)relstat.stat(0));
	return relstat;
    }

    void prepareNTerms(Term t1, Term t2, BitSetPlus mask) {
	if (nterm1 == null) {
	    nterm1 = new Term(t1);
	    nterm2 = new Term(t2);
	} else {   
	    nterm1.copy(t1);
	    nterm2.copy(t2);
	}
	for (int i = 0; i < nterm1.numAttrs(); i++) {
	    if (mask.get(i)) {
		nterm1.setVal(Term.RANGE, i);
		nterm2.setVal(Term.RANGE, i);
	    }
	}
    }

    AllDiffResult getSummarizedExceptions(Term t1, Term t2, BitSetPlus mask, Relation rel) throws Exception {
	prepareNTerms(t1,t2,mask);
	TermDesc tdesc = MdUtils.getTermDescForRange(nterm1);
	IterGen itergen = new IterGenC((RelCube)cube,nterm1,nterm2,tdesc);
        DiffParams diffPars = new DiffParams();
        diffPars._N = pars._N;
	diffPars._maxR = pars._maxR;
	if (pars._debug)
	    diffPars.setDebugOn();
	rel.setDiffParams(diffPars);
        FindDiff fd = new FindDiff(itergen,diffPars,null);
        return fd.allFinalResults();
    }


    void printBitList(Vector genvec) {
	for(int j=0;j<genvec.size();++j)
	    System.out.print(genvec.elementAt(j)+ " ");
	System.out.println(); 	
    }
    BitSetPlus join(BitSetPlus b1, BitSetPlus b2) {
	tmp.copy(b1);
	tmp.and(b2);
	if (tmp.numOnes() != b1.numOnes()-1)
	    return null;
	BitSetPlus v = new BitSetPlus();
	v.or(b2); v.or(b1);
//	System.out.println("Join: " + b1 + " & " + b2 + " = " + v);
	return v;
    }
    
    Vector getCandidates(Vector Lk) {
	Vector candGens = new Vector();
	for (int i = 0; i < Lk.size(); i++) {
	    for (int j = i + 1; j < Lk.size(); j++) {

		BitSetPlus v = join((BitSetPlus)Lk.elementAt(i), (BitSetPlus)Lk.elementAt(j));

		if ((v != null) && ! contains(candGens,v))
		    candGens.addElement(v);
	    }
	}
	return candGens;
    }
    boolean contains(Vector l, BitSetPlus v) {
	for (int i = 0; i < l.size(); i++) {
	    if (v.equals((BitSetPlus)l.elementAt(i)))
		return true;
	}
	return false;
    }

    int[] getBestPartition(int G,int size, AllDiffResult[] r){
	state=new int[G];
	int part[]=new int[G];
	mincost=100000;
	genPartitions(G,size,G,r,0,0,part);
	return state;
    }	
/*    Each generalization has answer size N along with its exceptions. The N rows of the answer need to be optimally 
    distributed among the G generalizations such that the total cost of the answer is minimum 
   
    g -- number of gen to be assigned slots
    k -- remaining slots
    G -- total number of generalizations
    partcost -- cost of previous partial solution (one additional slot of part is not assigned) 
    cost -- cost of partial solution
    part[] -- partial solution */

    void genPartitions(int g, int k, int G, AllDiffResult result[], double partcost, double cost,int part[]) { 
	if ((g==1)){
	   if (k>pars._N) part[G-1]=pars._N;
	   else part[G-1]=k;	
	   cost=partcost+result[G-1].getResult(k).cost();
	  //System.out.println("cost: "+cost);
	   for (int j=0; j<G; j++)
	       System.out.print(part[j]+" "); 
	   System.out.println(": "+cost);
	   if (cost==mincost) {// get thr min answer size for the same cost 
	      if (entropy(part)>entropy(state))
		 System.arraycopy(part,0,state,0,G);
	   }
	   if (cost<mincost) { 
		mincost = cost;
		System.out.println("mincost="+mincost);
		System.arraycopy(part,0,state,0,G);
	   }
	 }	
	
	else{
	   if (g==G) {cost=0; partcost=0; }
	   for (int i=1; i<=(pars.resultSize()-G+1); i++){
		if ((k-i)>0) {
		   if (i>pars._N) part[G-g]=pars._N;
		   else part[G-g]=i;
		   partcost = cost;
		   cost=cost+result[G-g].getResult(part[G-g]).cost();
		   genPartitions(g-1,k-i,G,result,partcost,cost,part);
		   cost=partcost;
		}
		else break;
	   }
	}
    }

    static double entropy(int st[]){
	int sum=0;
	double ent=0;
	double p;
	for (int i=0; i<st.length; i++)
	    sum+=st[i];
	for (int i=0; i<st.length; i++) {
	    p=(double)st[i]/sum;
	    ent+= -1*p*Math.log(p)/Math.log(2);
	}
//	System.out.println("entropy="+ent);
	return ent;
    }
   

    // 0 -- provider
    // 1 -- cubname
    // 2 -- usr name
    // 3 -- usr passwd
    // 4 -- params
    // 5,6,7 -- relax tuples
    // e.g. java relax.Relax Oracle data/pp icube icube "rel=boolean;thres=0.2" "Product;Geography;Platform;Year" "HRM/Payroll;United States;Single-User Other;1993" "HRM/Payroll;United States;Single-User Other;1994"

    public static void main(String args[]) {
	try {
	    Cube cube = MdUtils.parseInputPars(args);
	    RelaxParams pars = new RelaxParams(args[4]);
//	    RelRelax rel = new RelRelax((RelCube)cube, pars);
	    Relax rel = new Relax(cube,pars);
	    TermDesc tdesc = MdUtils.getColDesc(cube);
	    MTerm t1 = new MTerm(MdUtils.getTermMapping(cube,tdesc,args[5],args[6])); 
	    MTerm t2 = new MTerm(MdUtils.getTermMapping(cube,tdesc,args[5],args[7])); 
	    rel.relax(t1,t2,pars.getRel()); 
	} catch(Exception e) {
	    e.printStackTrace();
	}
    }
};
    
