/*
MdUtils : Provides some generic utilities for manipulating Terms
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package util;
import java.util.*;
import java.io.*;
import cube.*;

public class MdUtils {
    public static TermDesc getColDesc(Cube cube) {
	TermDesc coldesc = new TermDesc(cube.numAttrs());
	for (int i = 0, r = 0; i < cube.numDims(); i++) {
	    for (int j = cube.numLevels(i)-1; j >= 0; j--) {
		coldesc.desc(r++).init(i,j,cube.attrName(i,j));
	    }
	}
	return coldesc;
    }
    
    static int attrIndex(TermDesc coldesc, AttrDesc attrDesc) {
	for (int i = 0; i < coldesc.size(); i++)
	    if ((coldesc.desc(i).dim() == attrDesc.dim())
		&& (coldesc.desc(i).level() == attrDesc.level()))
		return i;
	return -1;
    }

    public static int attrIndex(TermDesc coldesc, int dim, int level) {
	for (int i = 0; i < coldesc.size(); i++)
	    if ((coldesc.desc(i).dim() == dim)
		&& (coldesc.desc(i).level() == level))
		return i;
	return -1;
    }
    
    public static Term getTermMapping(Cube cube, TermDesc coldesc, String[] attrName, String[] attrVal) throws Exception
    {
	Term t = new Term(coldesc);
	for (int i = 0; i < attrName.length; i++) {
	    getSingleTermMapping(t, cube, coldesc, attrName[i], attrVal[i]);
	}
	t.validate(cube);
	return t;
    }
    /*
    public static Term getMSOLAPTermMapping(Cube cube, TermDesc coldesc, String[] attrName, String[] attrVal) throws Exception
    {
	Term t = new Term(coldesc);
	for (int i = 0; i < attrName.length; i++) {
	    StringTokenizer dtok = new StringTokenizer(attrName[i],"[].");
	    int dim = dimIndex(cube,dtok.nextToken());
	    StringTokenizer tok = new StringTokenizer(attrVal[i],"]");
	    int lvl = cube.numLevels(dim) - (tok.countTokens()-2);
		if (lvl == cube.numLevels(dim))
			lvl--;
	    int attrind = attrIndex(coldesc,dim,lvl);
	    t.setVal(cube.memberIdFromName(dim,lvl,attrVal[i]),attrind); 
	}
	t.validate(cube);
	return t;
    }
    */
    public static util.AttrDesc attrDesc(Cube c, String attrName) {
	try {
	    for (int i = 0; i < c.numDims(); i++) {
		for (int j = 0; j < c.numLevels(i); j++) {
		    if (c.attrName(i,j).equals(attrName)) {
			util.AttrDesc adesc =  new util.AttrDesc();
			adesc.init(i,j,attrName);
			return adesc;
		    }
		}
	    }
	}catch(java.lang.NullPointerException ex4) {ex4.printStackTrace(); }
	return null;
    }
    public static void getSingleTermMapping(Term t, Cube cube, TermDesc coldesc, String attrName, String attrVal) throws DataCubeException {
	AttrDesc attrdesc = attrDesc(cube,attrName);
	if (attrdesc == null)
	    throw new DataCubeException("Invalid attribute Name: " + attrName);
	int attrInd = attrIndex(coldesc, attrdesc);
	if ((attrVal == null) )
	    t.setVal(Term.AGGREGATED, attrInd);
	else if (attrVal.equals("*"))
	    t.setVal(Term.RANGE, attrInd);
	else {
	    t.setVal(cube.memberIdFromName(attrdesc.dim(), attrdesc.level(), attrVal), attrInd);
	    t.setString(attrInd, attrVal);
	}
    }

    public static Term getTermMapping(Cube cube, TermDesc coldesc, Vector attrName, Vector attrVal) throws DataCubeException
    {
	Term t = new Term(coldesc);
	for (int i = 0; i < attrName.size(); i++) {
	    getSingleTermMapping(t, cube, coldesc,(String)attrName.elementAt(i), attrVal.elementAt(i).toString());
	}
	return t;
    }


    public static MTerm getMTermMapping(Cube cube, TermDesc coldesc, Vector attrName, Vector attrVal) throws DataCubeException
    {
	MTerm t = new MTerm(coldesc);
	if (attrName.size() == 0)
	    return t;
	t.setMeasure(Float.valueOf(attrVal.elementAt(0).toString()).floatValue());
	for (int i = 1; i < attrName.size(); i++) {
	    getSingleTermMapping(t, cube, coldesc,(String)attrName.elementAt(i), (String)attrVal.elementAt(i));
	}
	return t;
    }


    public static Term getTermMapping(Cube cube, TermDesc coldesc, String attrNames, String attrVals) throws Exception
    {
	return getTermMapping(cube,coldesc,MdUtils.getArray(attrNames),MdUtils.getArray(attrVals));
    }

    public static String[] getArray(String attrNames) {
	StringTokenizer nameTok = new StringTokenizer(attrNames, ";");
	String[] nameArray = new String[nameTok.countTokens()];
	for (int i = 0; nameTok.hasMoreTokens(); i++) {
	    nameArray[i] = nameTok.nextToken();
	}
	return nameArray;
    }

    public static Cube newCube(String provider, String cubeName, String usrName, String passwd, boolean debugdb) throws Exception {
	Cube cube;
	if (provider.startsWith("MSOLAP")) {
	    cube = (Cube)Class.forName("icube.cube.MDCubeC").newInstance();
	    cube.init("provider="+provider, cubeName);
	    // cube = new MDCubeC("provider="+provider, cubeName);
		   System.out.println("Fully qualified name " + cube.getClass().getName());
	} else { 
	    String relFile;
	    if (provider.equalsIgnoreCase("DB2"))
		relFile = cubeName + ".db2";
	    else
		relFile = cubeName + ".ora";
	    cube = new RelCubeC(cubeName + ".cub", relFile, usrName, passwd,debugdb);
	}
	return cube;
    }

    public static Cube parseInputPars(String argv[]) throws Exception
    { 
	Params pars = new Params(argv[4]);
	String password = argv[3];
	if (password.equals("?")) {
	    // prompt for password.
	    System.out.println("Password: ");
	    DataInputStream din = new DataInputStream(System.in);
	    password = din.readLine();
	}
	return newCube(argv[0], argv[1], argv[2], password,pars.debugQuery());
    }

    public static TermDesc getTermDescForRange(Term t1)
    {
	//
	// create a new term descriptor containing only the dimensions 
	// of the subcube.
	int size = 0;
	for (int i =0; i < t1.numAttrs(); i++)
	    if (t1.range(i))
		size++;
	if (size == 0)
	    return null;
	TermDesc tdesc = new TermDesc(size);
	size = 0;
	for (int i =0; i < t1.numAttrs(); i++)
	    if (t1.range(i))
		tdesc.desc(size++).init(t1.desc(i));
	return tdesc;
    }

    public static int[] defaultArray(Term t) {
	int[] array = new int[t.numAttrs()];
	int pos = 0;
	for (int i = 0; i < array.length; i++) {
	    if (t.range(i))
		array[pos++] = i;
	}
	if (pos < array.length-1)
	    array[pos] = -1;
	return array;
    }
}

