/*
MergeDataIter : Merges 2 sets of CubeRows 
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package util;

public class MergeDataIter implements Iter {
    CubeRows r1, r2;
    Object col1[] = null;
    Object col2[] = null;
    boolean forwarded1, forwarded2;
    protected ETerm tup;
    int ntups = 0;
    Term t1;
    public MergeDataIter(CubeRows s1, CubeRows s2,TermDesc tdesc, Term t) {
//	System.out.println("in mergeDataIter");
	r1 = s1; r2 = s2;
	forwarded1 = false;
	forwarded2 = false;
	t1 = t;
	tup = new ETerm(tdesc);
	col1 = new Object[numAttrs()+1];
	col2 = new Object[numAttrs()+1];
	
	while (moreTuples()) {
		System.out.println(getTuple().toString());
	}
    }
    public int numAttrs() {return tup.numAttrs();}
    public boolean moreTuples() {
	try {
	    if (r1 == null)
		return false;
	    boolean retval = next();
	    if (retval == false)
		close();
	    return retval;
	}  catch (Exception e) {
	    e.printStackTrace();
	}
	return false;
    }
    public ETerm getTuple() {
	return tup;
    }
    public void getTuple(ETerm thisTup) {
	thisTup.copy(tup);
	ntups++;
    }
    public TermDesc termDesc() {return tup.termDescr();}
    public void close() {
	try {
	    if (r1 != null) {
		System.out.println("MergeDataIter closed: ");
		r1.close();
		r2.close();
	    }
	    r1 = null;
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }
    protected void finalize() throws Throwable
    {
	close();
    }

    public int numColumns() {return r1.numColumns()+1;}


    void setVal(Object[] colm, Object[] colm1, Object[] colm2)
    {
	for (int i = 0; i < numAttrs(); i++) {
	    tup.setVal(colm[i],i);
	}
	double val1 = 0, val2 = 0;	
	Double v=new Double(0);
	if (colm1 != null) {		
	    val1=(v.valueOf(colm1[colm1.length-1].toString())).doubleValue();
	}	if (colm2 != null) {		
	    val2=(v.valueOf(colm2[colm2.length-1].toString())).doubleValue();	
	}

	tup.setMeasures(val1,val2);
    }

    void readRow(CubeRows cr, Object[] cols) throws Exception{
	int i = 0;
	for (int colIndex = 0; colIndex < cr.numColumns()-1; colIndex++) {
	    if (t1.range(colIndex)) 
		cols[i++] = cr.getObject(colIndex+1);	
	    else cr.getObject(colIndex+1);
        }
	cols[numAttrs()] = cr.getObject(cr.numColumns());
    }	
    
    int compare(Object[] col1, Object[] col2){
	for (int i = 0; i < col1.length-1; i++) {	
	    int comp = col1[i].toString().compareTo(col2[i].toString());
	    if (comp != 0)
		return comp;
	}
	return 0;
    }

    public boolean next() throws Exception{
	boolean retval1 = true;
	boolean retval2 = true;
	if (!forwarded1) {
	    retval1 = r1.next();
	    if (retval1)
		readRow(r1,col1);
	}
	if (!forwarded2) {
	    retval2 = r2.next();	
	    if (retval2)
		readRow(r2,col2);
	}
	if (!retval1 && !retval2)
	    return false;
	forwarded2 = false;
	forwarded1 = false;		

	if (retval1 == false) {
	    setVal(col2, null, col2);
	    return true;
	}	
	if (retval2 == false) {		
	    setVal(col1, col1, null);
	    return true;
	} 
	
	int comp = compare(col1, col2);
	if (comp == 0) {
	    setVal(col1, col1, col2);
	    return true;
	}
	if (comp < 0) {
	    setVal(col1, col1, null);
	    forwarded2 = true;
	    return true;
	}		
	setVal(col2, null, col2);
	forwarded1 = true;
	return true;
    }
}





