/*
Term : 
 Assume that all terms contain all attributes at all levels of the 
 hierarchy and in the same order
 A term is  thus a fixed length set of values 

 A term consists of (attribute, value) pairs where attribute refers to a
 level of a cube dimension and value can be one of:
     1. (All) -- aggregated to (All). This is the default
     2. *     -- range over all possible values
     3. constant -- a value from the domain.

    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package util;
import java.util.*;
import cube.*;



public class Term  implements java.io.Serializable  {
    public static final int SUBSUMES = 0;
    public static final int IS_SUBSUMED = 1;
    public static final int NO_OVERLAP = 2;
    public static final int DISJOINT = -1;
    public static final int EQUAL = 3;
    public static final int LESS_DETAILED = 4;
    public static final int MORE_DETAILED = 5;
    public static final int NONSEP_LESS_DETAILED = 6;
    public static final int NONSEP_MORE_DETAILED = 7;
    public static final int COVERS = 8;
    public static final int IS_COVERED = 9;
    
    public static final Integer CONSTANT = new Integer(1);
    public static final Integer RANGE = new Integer(-1);
    public static final Integer AGGREGATED = new Integer(-2);
    public static final Integer INVALID = new Integer(-3);

    TermDesc termDesc;  // meta data about the term. 
    public class AttrVal implements java.io.Serializable {
	Object val;
	public AttrVal() {
	    val = AGGREGATED;
	}
	public boolean range() {
	    if (val == RANGE)
		return true;
	    if (val.toString().equals("-1"))

	    // if (((Number)val).intValue() == -1)
		return true;
	    return false;
	}
	public boolean aggregated() {
	    if (val == AGGREGATED)
		return true;
	    return  (val.toString().equals("-2"));
		
	    // return (AGGREGATED.equals(val));
	}
	public boolean constant() {
	    return !range() && !aggregated() && !invalid();
	}
	public boolean invalid() {
	    return (val == INVALID);
	}
	public boolean same(AttrVal v) {
	    if (v == val)
		return true;
	  //  return v.toString().equals(val.toString());
	 return (val.equals(v.val));
	}
	public Object getVal() {
	    return val;
	    //	    return ((Integer)val).intValue();
	}
	
	public int intVal() {
	    return ((java.math.BigDecimal)val).intValue();
	}
	
	public void setVal(Object v) {
	    val = v;
	} 
	public void setVal(AttrVal v) {
	    val = v.val;
	}
	public Object getHandle() {return val;}
	public String toString() {
	    if (aggregated())
		return "(All)";
	    if (range())
		return "*";
	    return String.valueOf(val);
	}
	public String getString() {
	    return toString();
	}
	public void setString(String str) {
	    val = str;
	}
	public void setHandle(Object val1) {
	    val = val1;
	}
	public void fromString(String str) {
	    if (str.trim().equals("*"))
		setVal(Term.RANGE);
	    else if (str.equals("(All)"))
		setVal(AGGREGATED);
	    else setVal(Integer.valueOf(str));
	}
	public void fromString(String str,Cube c,AttrDesc attrdesc) {
	    if (str.trim().equals("*"))
		setVal(Term.RANGE);
	    else if (str.equals("(All)"))
		setVal(AGGREGATED);
	    else {
		try {
		    setVal(c.memberIdFromName(attrdesc.dim(),attrdesc.level(), str));
		} catch (Exception e) {
		    e.printStackTrace();
		    System.exit(0);
		}
	    }
	}
    }
    AttrVal cols[] = null;

    private void initCols() {
	int ncols = 0;
	if (termDesc != null) {
	    ncols = termDesc.size();
	    cols = new AttrVal[ncols];
	}
	for (int i = 0; i < ncols; i++) {
	    cols[i] = new AttrVal();
	}
    }
    public Term(TermDesc tdesc) {
	termDesc = tdesc;
	initCols();
    }
    public Term(Term t) {
	copy(t);
    }
    public int numAttrs() {return (cols == null)?0:cols.length;}
    public void copy(Term t) {
	if ((cols == null) || (cols.length < t.numAttrs())) {
	    cols = new AttrVal[t.numAttrs()];
	    for (int i = 0; i < cols.length; i++) {
		cols[i] = new AttrVal();
	    }
	    termDesc = t.termDesc;
	}
	for (int i = 0; i < t.numAttrs(); i++) {
	    cols[i].setVal(t.attrVal(i));
	}
    }
    public boolean aggregated(Object val) {
	if (val == Term.AGGREGATED)
	    return true;
	if ((val.getClass() == String.class) && val.equals("(All)"))
	    return true;
	return false;
    }
    public AttrDesc desc(int attrIndex) {return termDesc.desc(attrIndex);}
    public TermDesc termDescr() {return termDesc;}
    public int size() { return (termDesc == null)?0:termDesc.size();}
    public void setVal(Object v, int i) {
	cols[i].setVal(v);
    }
    public void setVal(String attrName, int i) {
	setString(i, attrName);
    }
    public void setVal(AttrVal v, int i) {
	cols[i].setVal(v);
    }
    public boolean isEqual(Term t, BitSet mask) {
	if (t.cols.length != cols.length) 
	    return false;
	for (int i = 0; i < cols.length; i++) {
	    if ((mask != null) && !mask.get(i)) continue;
	    if (! cols[i].same(t.attrVal(i)))
		return false;
	}
	return true;
    }
    //	    
    // 
    // a is LESS_DETAILED than b if for all i ai = bi or ai = all
    // Coverage of a could be more than b or same as b.
    public int detailRel(Term t) {
	if (t.cols.length != cols.length) 
	    return DISJOINT; 
	int rel = EQUAL;
	for (int i = 0; i < cols.length; i++) {
	    if (cols[i].same(t.attrVal(i)))
		continue;
	    if (cols[i].constant() && t.attrVal(i).constant())
    		return NO_OVERLAP;
	    if (cols[i].aggregated())
		if (rel != MORE_DETAILED) {
		    rel = LESS_DETAILED;
		} else
		    return DISJOINT;
	    if (t.attrVal(i).aggregated())
		if (rel != LESS_DETAILED) {
		    rel = MORE_DETAILED;
		} else
		    return DISJOINT;
	}
	return rel;
    }
    // check if t can be seperated from this term without subtraction..
    public boolean seperable(Term t) {
	for (int i = 0; i < cols.length; i++) {
	    if (cols[i].aggregated() && t.attrVal(i).constant())
		return false;
	}
	return true;
    }
    public boolean lessDetailNonSeperable(Term t) {
	int rel = detailRel(t);
	if (rel != LESS_DETAILED)
	    return false;
	return !seperable(t);
    }
    public int subsumeRel(Term t, BitSet mask) {
	if (t.cols.length != cols.length) 
	    return DISJOINT; 
	int rel = EQUAL;
	for (int i = 0; i < cols.length; i++) {
	    if ((mask != null) && !mask.get(i)) continue;
	    if (cols[i].same(t.attrVal(i)))
		continue;
	    if (cols[i].range())
		if (rel != IS_SUBSUMED) {
		    rel = SUBSUMES;
		    continue;
		} else
		    return DISJOINT;
	    if (t.attrVal(i).range())
		if (rel != SUBSUMES) {
		    rel = IS_SUBSUMED;
		    continue;
		} else
		    return DISJOINT;
	    return DISJOINT;
	}
	return rel;	
    }

    public int coverRel(Term t, BitSet mask) {
	if (t.cols.length != cols.length) 
	    return NO_OVERLAP; 
	int rel = EQUAL;
	for (int i = 0; i < cols.length; i++) {
	    if ((mask != null) && !mask.get(i)) continue;
	    if (cols[i].same(t.attrVal(i)))
		continue;
	    if (cols[i].constant() && t.attrVal(i).constant())
		return NO_OVERLAP;
	    if (cols[i].constant())
		if (rel != COVERS) {
		    rel = IS_COVERED;
		    continue;
		} else
		    return DISJOINT;
	    if (t.attrVal(i).constant())
		if (rel != IS_COVERED) {
		    rel = COVERS;
		    continue;
		} else
		    return DISJOINT;
	    return DISJOINT;
	}
	return rel;	
    }

    public boolean constant(int i) {
	return cols[i].constant();
    }
    public boolean range(int i) {
	return cols[i].range();
    }
    public boolean aggregated(int i) {
	return cols[i].aggregated();
    }
    public AttrVal attrVal(int i) {
	return cols[i];
    }
    public Object val(int i) {
	return cols[i].getVal();
    }

    public int intVal(int i) {
	return cols[i].intVal();
    }
    public String toString() {
	String st = "";
	if (cols == null) return st;
	for (int i = 0; i < cols.length; i++) {
	    if (i > 0) st += ", ";
	    st += attrVal(i).toString();
	}
	return st;
    }
    public String toFullString() {
	String st = "";
	if (cols == null) return st;
	for (int i = 0; i < cols.length; i++) {
	    if (i > 0) st += ", ";
	    st += (attrVal(i).toString() + getString(i));
	}
	return st;
    }
    public void fromString(String tstr) {
	StringTokenizer tok = new StringTokenizer(tstr, ", ");
	for (int ind = 0; tok.hasMoreTokens() && (ind < cols.length); ind++) {
	    attrVal(ind).fromString(tok.nextToken());
	}
    }
    public void fromString(String tstr,Cube c) {
	StringTokenizer tok = new StringTokenizer(tstr, ",");
	for (int ind = 0; tok.hasMoreTokens() && (ind < cols.length); ind++) {
	    String strAttr = tok.nextToken();
	    attrVal(ind).fromString(strAttr,c,desc(ind));
	    setString(ind, strAttr);
	}
    }

    public  String toMetaString() {return termDesc.toMetaString();}
    public void print() {
	System.out.println(toString());
    }
    public String getExternalName(Cube c, int dd) {
	try {
	    return c.memberNameFromId(desc(dd).dim(), desc(dd).level(), cols[dd].getVal());
	} catch (Exception e) {
	    e.printStackTrace();
	}
	return null;
    }
    public void setExternalNames(Cube c) {
	for (int i = 0; i < numAttrs(); i++) {
	    if (constant(i))
		setString(i, getExternalName(c, i));
	}
    } 
    public String getString(int i) {
	return cols[i].getString();
    }
    public void setString(int i, String val) {
	cols[i].setString(val);
    }
    /*
    public void copyExternalNames(Term t) {
	for (int i = 0; i < numAttrs(); i++) {
	    setString(i, t.getString(i));
	}
	} */
    public void mask(BitSet msk) {
	for (int i = 0; i < cols.length; i++)
	    if (!msk.get(i))
		cols[i].setVal(AGGREGATED);
    }
    public int numNonAgg() {
	int cnt = 0;
	if (cols == null)
	    return 0;
	for (int i = 0; i < cols.length; i++)
	    if (cols[i].aggregated())
		cnt++;
	return cols.length - cnt;
    }
    public void setCombinedTerms(Term t1, Term t2) {
	// merge t1 and t2 to contain the greater of the details but lesser of the covereage 
	// of these two terms..
	for (int i = 0; i < t1.numAttrs(); i++) {
	    setVal(t1.attrVal(i), i);
	    if (t1.attrVal(i).same(t2.attrVal(i)) || t2.aggregated(i) || t1.constant(i))
		continue;
	    setVal(t2.attrVal(i),i);
	}
    }
    public boolean elementOf(Vector termVec)
    {
	for (int i = 0; i < termVec.size(); i++) {
	    if (termVec == termVec.elementAt(i))
		return true;
	    if (isEqual((Term)termVec.elementAt(i),null))
		return true;
	}
	return false;
    }

    public boolean valid() {
	for (int i = 0; i < numAttrs(); i++) {
	    if (desc(i).level() == 0) continue;
	    if (constant(i)) continue;
	    for (int j = 0; j < numAttrs(); j++) {
		if (desc(j).dim() != desc(i).dim()) continue;
		if (desc(j).level() >= desc(i).level()) continue;
		if (range(j) && aggregated(i))
		    return false;
		if (constant(j) && !constant(i))
		    return false;
	    }
	}
	return true;
    }
    //
    // If a lower level attribute is "*", make all higher level attributes 
    // the same. Similarly if it is a constant, update the higher level 
    // attributes with the right parent value
    public void validate(Cube cube) throws Exception {
	for (int i = 0; i < numAttrs(); i++) {
	    if (constant(i))
		setVal(cube.memberIdFromName(desc(i).dim(), desc(i).level(), attrVal(i).getString()), i); 
	}
	for (int i = 0; i < numAttrs(); i++) {
	    if (desc(i).level() == 0) continue;
	    if (constant(i)) continue;
	    for (int j = 0; j < numAttrs(); j++) {
		if (desc(j).dim() != desc(i).dim()) continue;
		if (desc(j).level() >= desc(i).level()) continue;
		if (range(j) && aggregated(i))
		    setVal(Term.RANGE, i);
		if (constant(j) && !constant(i)) {
		    setVal(cube.getParent(desc(j).dim(),desc(j).level(), cols[j].getVal(), desc(i).level()), i);
		}
		// throw new Exception("Parent attribute not constant whereas child is a constant");
	    }
	}
    }
}
